import os
import sys
from pathlib import Path

from inspector_commons.context import Context as CommonContext  # type: ignore
from inspector_commons.utils import IS_WINDOWS  # type: ignore

from inspector.windows.base import WindowState  # type: ignore


class Context(CommonContext):
    def create_window(self, kind, factory, **kwargs):
        if kind == "windows" and not IS_WINDOWS:
            raise NotImplementedError(
                "Windows locators are not available on non Windows platforms"
            )

        self.logger.debug("Factory Window type: %s", factory)
        if factory is None:
            raise KeyError(f"Unknown window type: {kind}")

        self.logger.debug("Created instances: %s", self.INSTANCES)
        window = self.INSTANCES[kind]
        if window is not None and window.is_valid:
            self.logger.debug("Restoring window...")
            window.restore()
            return window

        window = factory.create(self, **kwargs)
        self.INSTANCES[kind] = window
        self.logger.debug("Instances: %s", self.INSTANCES)
        self.logger.debug("Created window: %s", window)

        self.windows.append(window)
        return window

    def close_windows(self):
        self.logger.debug("Closing windows: %s", self.windows)
        for window in reversed(self.windows):
            if window.state != WindowState.CLOSED:
                self.logger.debug("Destroying window: %s", window.uid)
                window.destroy()
            else:
                self.logger.debug("Window already closed: %s", window.uid)

    # NOTE: need to overwrite the entrypoint as the Paths from the package
    # will refer to the package itself
    @property
    def entrypoint(self):
        # NOTE: pywebview uses sys.argv[0] as base
        base = Path(sys.argv[0]).resolve().parent
        static = Path(__file__).resolve().parent.parent / "static"
        return os.path.relpath(str(static), str(base))



from inspector_commons.bridge.bridge_browser import BrowserBridge  # type: ignore
from inspector.windows.base import Window


class BrowserWindow(Window):
    BRIDGE = BrowserBridge
    DEFAULTS = {
        "title": "Robocorp Web Element Locator",
        "url": "browser.html",
        "width": 480,
        "height": 100,
        "on_top": True,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._force_closing = False

    def on_closing(self):
        super().on_closing()

        driver = self._context.webdriver
        if driver is not None and driver.is_running:
            try:
                driver.clear()
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.debug("Failed to clear webdriver: %s", exc)

        return True



# pylint: disable=too-many-instance-attributes
import base64
import json
import logging
import traceback
from io import BytesIO

import tkinter as tk  # type: ignore
import mss  # type: ignore
from PIL import Image, ImageTk  # type: ignore


class SnippingTool:
    def __init__(self, timeout=30):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.timeout = timeout

        #: Final result
        self.error = None
        self.result = None

        with mss.mss() as sct:
            # NB: Always uses the left-most monitor
            monitor = sct.monitors[1]
            screenshot = sct.grab(monitor)

        #: Display dimensions
        self.width = monitor["width"]
        self.height = monitor["height"]

        self.logger.debug("Screen dimensions: %dx%d", self.width, self.height)
        self.logger.debug("Screenshot size: %dx%d", *screenshot.size)

        #: Desktop screenshot (as PIL.Image)
        self.screenshot_image = Image.frombytes(
            "RGB", screenshot.size, screenshot.bgra, "raw", "BGRX"
        )

        #: Desktop screenshot (as PNG bytes)
        self.screenshot_bytes = mss.tools.to_png(screenshot.rgb, screenshot.size)

        #: Current snip coordinates
        self.start_x = 0
        self.start_y = 0
        self.end_x = 0
        self.end_y = 0

        #: Snipping outline rectangle
        self.outline = None

        # Root widget
        self.root = tk.Tk()
        self.root.configure(bg="#f0f0f0")

        # Bring window to full screen and top most level. For some reason this
        # overrideredirect toggling is required to make this work on macOS:
        # https://stackoverflow.com/a/42173885/6734941
        try:
            self.root.overrideredirect(True)
            self.root.overrideredirect(False)
        except Exception:  # pylint: disable=broad-except
            self.logger.warning(traceback.format_exc())

        self.root.attributes("-fullscreen", True)
        self.root.attributes("-topmost", True)

        # Create canvas for drawing content
        self.canvas = tk.Canvas(
            self.root,
            width=self.width,
            height=self.height,
            cursor="crosshair",
        )

        # Use screenshot as background, resized to fit screen in case of scaling
        self.background = ImageTk.PhotoImage(
            self.screenshot_image.resize((self.width, self.height), Image.LANCZOS)
        )

        self.canvas.create_image((0, 0), image=self.background, anchor="nw")
        self.canvas.pack()

        # Focus users input on the canvas
        # FIXME: Doesn't work at least on macOS
        self.canvas.focus()
        self.canvas.focus_force()

        # Connect the event handlers
        self.canvas.bind("<Escape>", self._on_escape)
        self.canvas.bind("<B1-Motion>", self._on_move)
        self.canvas.bind("<ButtonPress-1>", self._on_button_press)
        self.canvas.bind("<ButtonRelease-1>", self._on_button_release)

    def run(self):
        # self.logger.debug("Starting application")
        self.root.after(int(self.timeout * 1000), self._on_timeout)
        self.root.mainloop()

        if self.error:
            raise RuntimeError(self.error)

        return self.result

    def quit(self):
        self.root.quit()

    def _to_width(self, value):
        return max(min(value, self.width), 0)

    def _to_height(self, value):
        return max(min(value, self.width), 0)

    def _on_timeout(self, _):
        self.logger.warning("Timeout reached (%d seconds)", self.timeout)
        self.error = "Timeout reached"
        self.quit()

    def _on_escape(self, _):
        self.logger.warning("Aborted by user")
        self.error = "Aborted by user"
        self.quit()

    def _on_button_press(self, event):
        self.start_x = self.end_x = self._to_width(event.x)
        self.start_y = self.end_y = self._to_height(event.y)

        if not self.outline:
            self.outline = self.canvas.create_rectangle(
                self.start_x,
                self.start_y,
                self.end_x,
                self.end_y,
                outline="#1B97F3",
                stipple="gray12",
            )

    def _on_move(self, event):
        self.end_x = self._to_width(event.x)
        self.end_y = self._to_height(event.y)

        if self.outline:
            self.canvas.coords(
                self.outline, self.start_x, self.start_y, self.end_x, self.end_y
            )

    def _on_button_release(self, _):
        width, height = self.screenshot_image.size
        width_scale = width / self.root.winfo_width()
        height_scale = height / self.root.winfo_height()

        if width_scale != height_scale:
            # This might mean that the window is not truly fullscreen or
            # that the screenshot missed something
            self.logger.warning(
                "Uneven width/height scaling (%s / %s)", width_scale, height_scale
            )

        scale_factor = height_scale
        self.logger.debug("Calculated scale factor: %f", scale_factor)

        coordinates = (
            int(scale_factor * min(self.start_x, self.end_x)),
            int(scale_factor * min(self.start_y, self.end_y)),
            int(scale_factor * max(self.start_x, self.end_x)),
            int(scale_factor * max(self.start_y, self.end_y)),
        )

        self.logger.info("Snip coordinates: %s", coordinates)
        snip = self.screenshot_image.crop(coordinates)

        stream = BytesIO()
        snip.save(stream, format="png")
        snip_bytes = stream.getvalue()

        self.result = base64.b64encode(snip_bytes).decode()
        self.quit()


def main():
    log_datefmt = "%Y/%m/%d %H:%M:%S"
    log_format = "%(asctime)s.%(msecs)03d › %(levelname)s › %(name)s › %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=log_format, datefmt=log_datefmt)

    # Reduce unwanted debug logging
    logging.getLogger("PIL").setLevel(logging.INFO)

    try:
        snipper = SnippingTool()
        result = snipper.run()
        output = {"image": str(result)}
    except Exception as err:  # pylint: disable=broad-except
        logging.debug(traceback.format_exc())
        output = {"error": str(err)}

    print(json.dumps(output), flush=True)


if __name__ == "__main__":
    main()



from inspector_commons.utils import IS_WINDOWS  # type: ignore
from .web_recorder import WebRecorderWindow
from .base import WindowState
from .browser import BrowserWindow
from .manager import ManagerWindow
from .image import ImageWindow

WINDOWS = {
    "browser": BrowserWindow,
    "image": ImageWindow,
    "manager": ManagerWindow,
    "web-recorder": WebRecorderWindow,
}

if IS_WINDOWS:
    from .windows import WindowsWindow

    WINDOWS["windows"] = WindowsWindow



from inspector_commons.bridge.bridge_recorder import RecorderBridge  # type: ignore
from inspector.windows.browser import BrowserWindow


class WebRecorderWindow(BrowserWindow):
    BRIDGE = RecorderBridge
    DEFAULTS = {
        "title": "Robocorp Web Element Recorder",
        "url": "webrecorder.html",
        "width": 480,
        "height": 100,
        "on_top": True,
    }



import os
from abc import ABC
from enum import Enum, auto

import webview  # type: ignore

from inspector_commons.bridge.base import Bridge  # type: ignore


APP_WINDOW_TITLE = "UI Inspector"


class WindowState(Enum):
    CREATED = auto()
    SHOWN = auto()
    LOADED = auto()
    CLOSING = auto()
    CLOSED = auto()


class Window(ABC):
    BRIDGE = Bridge
    DEFAULTS = {
        "title": "Title",
        "url": "index.html",
        "width": 640,
        "height": 480,
    }

    @classmethod
    def create(cls, context, **kwargs):
        options = dict(cls.DEFAULTS)
        options.update(kwargs)
        options.setdefault("title", APP_WINDOW_TITLE)

        context.logger.info("Window Context: %s", context)
        context.logger.info("Entrypoint: %s", context.entrypoint)
        options["url"] = os.path.join(context.entrypoint, options["url"])

        context.logger.info("Options: %s", options)
        bridge = cls.BRIDGE(context)
        window = webview.create_window(
            js_api=bridge,
            min_size=(cls.DEFAULTS["width"], cls.DEFAULTS["height"]),
            resizable=False,
            **options,
        )
        instance = cls(context, window, bridge)
        bridge.set_window(instance)

        context.logger.info("Window Created: %s", window)
        return instance

    def __init__(self, context, window, bridge):
        self._context = context
        self._window = window
        self._bridge = bridge
        self._state = WindowState.CREATED

        # Attach event callbacks
        self._window.events.closed += self.on_closed
        self._window.events.closing += self.on_closing
        self._window.events.shown += self.on_shown
        self._window.events.loaded += self.on_loaded

        # MacOS closing flag
        self._force_closing = False

    def __getattr__(self, name):
        return getattr(self._window, name)

    @property
    def logger(self):
        return self._context.logger

    @property
    def state(self):
        return self._state

    @property
    def is_valid(self):
        return self.state not in (WindowState.CLOSING, WindowState.CLOSED)

    def on_shown(self):
        self.logger.debug("Window shown")
        self._state = WindowState.SHOWN

    def on_loaded(self):
        self.logger.debug("Window loaded")
        self._state = WindowState.LOADED

    def on_closing(self):
        self.logger.debug("Window closing")
        self._state = WindowState.CLOSING

    def on_closed(self):
        self.logger.debug("Window closed")
        self._state = WindowState.CLOSED



from pathlib import Path
from typing import List
from enum import Enum

from inspector_commons.utils import IS_WINDOWS  # type: ignore
from inspector_commons.bridge.mixin import traceback  # type: ignore
from inspector.windows.base import Bridge, Window
from inspector.windows.windows import WindowsWindow
from .web_recorder import WebRecorderWindow
from .browser import BrowserWindow
from .image import ImageWindow


class SupportedInstances(Enum):
    MANAGER = "manager"
    BROWSER = "browser"
    IMAGE = "image"
    WEBRECORDER = "web-recorder"
    WINDOWS = "windows"


class ManagerBridge(Bridge):
    """Javascript API bridge."""

    def path(self, length=32):
        # Return absolute path if short enough
        absolute = Path(self.ctx.database.path).resolve()
        if len(str(absolute)) <= length:
            return str(absolute)

        # Return partial path by removing parent directories
        parts = absolute.parts
        while parts:
            parts = parts[1:]
            shortened = str(Path("...", *parts))
            if len(shortened) <= length:
                return shortened

        # Return just filename or partial filename
        if len(absolute.name) > length:
            return "..." + absolute.name[-length:]
        else:
            return absolute.name

    def list(self):
        with self.ctx.database.lock:
            self.ctx.database.load()
            db_list = self.ctx.database.list()
            return db_list

    @traceback
    def rename(self, before, after):
        with self.ctx.database.lock:
            self.ctx.database.load()
            locator = self.ctx.database.pop(before)
            self.ctx.database.update(after, locator)

    @traceback
    def add(self, kind):
        self.ctx.selected = None
        self.ctx.create_window(kind, WINDOWS[kind])

    @traceback
    def edit(self, name):
        locator = self.ctx.load_locator(name)
        if not locator:
            self.logger.error("No locator with name: %s", name)
            return
        self.ctx.selected = name
        self.ctx.create_window(locator["type"], WINDOWS[locator["type"]])

    @traceback
    def remove(self, name):
        with self.ctx.database.lock:
            self.ctx.database.delete(name)

    @traceback
    def list_supported_locators(self) -> List[str]:  # pylint: disable=no-self-use
        if IS_WINDOWS:
            return [
                SupportedInstances.MANAGER.value,
                SupportedInstances.BROWSER.value,
                SupportedInstances.IMAGE.value,
                SupportedInstances.WEBRECORDER.value,
                SupportedInstances.WINDOWS.value,
            ]
        return [
            SupportedInstances.MANAGER.value,
            SupportedInstances.BROWSER.value,
            SupportedInstances.IMAGE.value,
            SupportedInstances.WEBRECORDER.value,
        ]


class ManagerWindow(Window):
    BRIDGE = ManagerBridge
    DEFAULTS = {
        "title": "Robocorp Inspector Manager",
        "url": "manager.html",
        "width": 450,
        "height": 900,
    }


WINDOWS = {
    "browser": BrowserWindow,
    "image": ImageWindow,
    "manager": ManagerWindow,
    "web-recorder": WebRecorderWindow,
    "windows": WindowsWindow,
}



from inspector_commons.bridge.bridge_windows import WindowsBridge  # type: ignore
from inspector.windows.base import Window


class WindowsWindow(Window):
    BRIDGE = WindowsBridge
    DEFAULTS = {
        "title": "Robocorp App Element Locator",
        "url": "windows.html",
        "width": 560,
        "height": 720,
    }



import logging
import os
import signal
import sys
import webview  # type: ignore
import psutil  # type: ignore

from inspector_commons.metric import MetricStart  # type: ignore
from inspector.windows.context import Context
from inspector.windows import WINDOWS

PYWEBVIEW_GUI = ["qt", "gtk", "cef", "mshtml", "edgechromium", "edgehtml"]


class App:
    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.ctx = Context(self.logger, config)
        self.process = None

    def start(self, root="manager"):
        assert len(self.ctx.windows) == 0, "Application already running"
        self.logger.info("Window will open for: %s", root)

        self.ctx.database.load()
        if self.ctx.database.error:
            self.logger.warning(*self.ctx.database.error)

        factory = WINDOWS.get(root)
        root = self.ctx.create_window(kind=root, factory=factory)
        root.closed += self.stop

        self.ctx.telemetry.send(MetricStart())

        webview.start(self._on_start, debug=self.ctx.is_debug)

    def stop(self):
        # kill all processes as chromedriver hangs
        self._kill_all_child_processes()
        # force exit application as there is nothing to worry about
        # and child processes hang
        os._exit(0)  # pylint: disable=protected-access

    def open(self, kind):
        self.start(kind)

    def edit(self, name):
        locator = self.ctx.load_locator(name)

        if locator is None:
            names = "\n".join(f" - {name}" for name in self.ctx.database.names)
            if names:
                self.logger.info("No locator with name: %s\n", name)
                self.logger.info("Possible options: %s\n", names)
            else:
                self.logger.error("No locators in database")
            sys.exit(1)

        self.ctx.selected = name
        self.start(locator["type"])

    def _on_start(self):
        self.logger.info("Starting root window...")
        if self.ctx.is_debug:
            # Print URL for automation
            root = self.ctx.windows[0]
            self.logger.info(root.get_current_url())

    def _kill_all_child_processes(self):
        self.logger.debug("Stopping all child processes...")
        for child in psutil.Process().children(recursive=True):
            try:
                self.logger.debug("Stopping PROCESS: %s", child.pid)
                os.kill(child.pid, signal.SIGTERM)
                child.kill()
                child.wait()
            except Exception:
                pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inspector', 'inspector.windows']

package_data = \
{'': ['*'], 'inspector': ['static/*', 'static/assets/fonts/*']}

install_requires = \
['clr-loader==0.1.7',
 'importlib-metadata==4.13.0',
 'mss>=6.1.0,<7.0.0',
 'psutil>=5.9.2,<6.0.0',
 'pynput-robocorp-fork>=4.0.0,<5.0.0',
 'pythonnet==3.0.0a2',
 'pywebview==3.6.3',
 'requests>=2.25.1,<3.0.0',
 'robocorp-inspector-commons>=0.5.3,<0.6.0',
 'rpaframework-core>=9.1.0,<10.0.0',
 'rpaframework-recognition>=4.0.2,<5.0.0',
 'tk-tools>=0.16.0,<0.17.0',
 'typing-extensions>=3.10.0,<4.0.0',
 'uiautomation>=2.0.16,<3.0.0']

extras_require = \
{':python_full_version != "3.7.6" and python_full_version != "3.8.1" and sys_platform == "win32"': ['pywin32>=302,<304'],
 ':sys_platform == "darwin"': ['pyobjc-core>=7.3,<8.0',
                               'pyobjc-framework-cocoa>=7.3,<8.0',
                               'pyobjc-framework-quartz>=7.3,<8.0',
                               'pyobjc-framework-webkit>=7.3,<8.0']}

entry_points = \
{'console_scripts': ['inspector = inspector.cli:run']}

setup_kwargs = {
    'name': 'robocorp-inspector',
    'version': '0.8.2',
    'description': 'Robocorp Inspector',
    'long_description': '# Robocorp Inspector\n\nRobocorp Inspector is a tool for exploring various user interfaces\nand developing ways to target elements within them. An expression\nthat can target specific UI elemements is called a _locator_, and\nthese locators can be used to automate applications typically\nused by humans.\n\n## Dependencies\n\nYou might need to create a `.npmrc` file at project level with contents similar to the following, but with your own `authToken`.\nThis is needed for private repositories.\n\n```\nregistry=https://registry.npmjs.org/\n@robocorp:registry=https://npm.pkg.github.com/\n//npm.pkg.github.com/:_authToken=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n```\n\n> There is a hard dependency to the [Inspector Commons](https://github.com/robocorp/inspector-commons) implementation.\n> Most of the implementation resides in `inspector-commons` and if you spot any misalignment then you should correct it\n\n## Development\n\nThe project uses `invoke` for overall project management, `poetry` for\npython dependencies and environments, and `yarn` for Javascript dependencies\nand building.\n\nBoth `invoke` and `poetry` should be installed via pip: `pip install poetry invoke`\n\n- To see all possible tasks: `invoke --list`\n- To run the project: `invoke run `\n- For a quick build and run you can try running: `inv build-js && inv build && inv run`\n- To clean the dev environment you can use `inv clean` or `inv clean --force`\n\nAll source code is hosted on [GitHub](https://github.com/robocorp/inspector/).\n\n### Python & NPM\n\nTo launch the development environment you\'ll need:\n```\npyenv + virtualenv -> these will help building a dedicated python virtual environment\nnvm -> will help with a contained version of node + npm\n```\n\nIn order for everything to install and build properly use the following versions:\n```\npython -> v3.8.10\nnode -> v16.14.2\nnpm -> 8.5.0\n```\n\n## Usage\n\nRobocorp Inspector is distributed as a Python package with all front-end\ncomponents compiled and included statically.\n\nIf the package (and all required dependencies) is installed manually,\nit can be run with the command: `inspector`.\n\n## Code Organization\n\n> Attention: these might change over time & hopefully they will be maintained.\n\n### Inspector Class Diagram\n\n- not extremely precise\n- created to show how things link together from local implementation to `inspector-commons`\n\n![Inspector Class Diagram](./assets/InspectorClassDiagram.jpg)\n\n---\n\n<p align="center">\n  <img height="100" src="https://cdn.robocorp.com/brand/Logo/Dark%20logo%20transparent%20with%20buffer%20space/Dark%20logo%20transparent%20with%20buffer%20space.svg">\n</p>\n',
    'author': 'Ossi Rajuvaara',
    'author_email': 'ossi@rajuvaara.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/robocorp/inspector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
